#include "Vhub75e.h"
#include "verilated.h"
#include "verilated_vcd_c.h"
#include <array>
#include <cstdint>
#include <memory>
#include <queue>
#include <stdio.h>
#include <vector>

// slices the RGB values for us.
uint8_t rgb_slice(uint32_t rgb, uint8_t bit) {
  if (bit > 8) {
    // todo: panic
    return 0;
  }
  uint8_t r = (rgb >> (16 + bit)) & 1;
  uint8_t g = (rgb >> (8 + bit)) & 1;
  uint8_t b = (rgb >> bit) & 1;
  return (r << 2) & (g << 1) & (b << 1);
}
class HUB75Reciever {

  typedef std::vector<unsigned char> row_array;
  int xsize;
  int ysize;

  row_array row_upper;
  row_array row_lower;

  // the previous row values that were latched in.
  std::vector<row_array> past_rows;
  // the pulse width for each output, in clock cycles.
  std::vector<int> pulse_widths;

  int bit_position = 7; // the bit that is currently being shifted in

  int output_period_cnt;
  // if oe = 0, count clocks. when oe = 1, store value into
  // pulse_widths[display_bit];

  // previous latch value, used to identify when to latch.
  unsigned char prev_latch = 0;
  // previous display clock value, used to detect rising edge.
  unsigned char prev_display_clk = 0;

  unsigned char prev_oe = 0;

public:
  HUB75Reciever(int xsize, int ysize) : row_upper(xsize, 0) {
    this->xsize = xsize;
    this->ysize = ysize;
  };

  // evaluates the reciever.
  void eval(const int oe, const int latch, const int display_clk,
            const int rgb0) {

    if (prev_display_clk == 0 && display_clk == 1) {
      // display clock rising edge.
      row_upper.push_back(rgb0);
    }

    if (prev_latch == 0 && latch == 1) {
      // latch in the data: reverse the rows, and pu
      std::reverse(row_upper.begin(), row_upper.end());
      past_rows.push_back(row_upper);
      row_upper.clear();
    }
    if (oe == 0) {
      if (prev_oe == 1) {
        // falling edge.
        output_period_cnt = 0;
      } else {
        output_period_cnt++;
      }
    } else {
      // rising edge: store the output
    }

    // update previous values
    prev_display_clk = display_clk;
    prev_latch = latch;
    prev_oe = oe;
  }
};

class FakeBRAM {
  std::array<uint64_t, 512> ram{};

  std::queue<uint16_t> addr_lookup_q;

public:
  FakeBRAM(int latency) : addr_lookup_q() {
    for (int i = 0; i < latency; i++) {
      addr_lookup_q.push(0);
    }

    for (int i = 0; i < ram.size(); i++) {
      ram[i] = i;
    }
  };

  uint64_t tick(int addr) {
    // we push and pop in the same tick: this way we keep the queue the same
    // size, acting as a pipeline delay.
    addr_lookup_q.push(addr);

    auto addr_to_load = addr_lookup_q.front();
    addr_lookup_q.pop();
    return ram.at(addr_to_load);
  }

  // TODO: allow accessing/setting the data
};

void LineDriverTest(VerilatedContext &ctx) {
  // create the hub75e driver and run some basic tests

  // we generate 512 random color values (24 bits)
  // we load them in.
  unsigned long counter = 0; // counts positive edges.
  unsigned long simtime = 0;
  auto dut = std::make_unique<Vhub75e>(&ctx, "dut");

  VerilatedVcdC *m_trace = new VerilatedVcdC;
  dut->trace(m_trace, 5);

  auto bram = FakeBRAM(1);
  printf("Performing basic test\n");
  bool done = false;
  m_trace->open("waveform.vcd");

  while (!done) {
    dut->clk ^= 1; // toggle clock
    dut->eval();

    if (dut->clk == 1) {
      // rising edge.
      counter++;
      dut->pixbuf_data = bram.tick(dut->pixbuf_addr);
    }

    dut->write_trig = counter < 20 ? 1 : 0;

    if (counter >= 250000) {
      done = true;
    }
    m_trace->dump(simtime);
    simtime++;
  }
  m_trace->close();
};

int main() {
  auto ctx = std::make_unique<VerilatedContext>();
  ctx->traceEverOn(true);
  printf("hello world!\n");
  LineDriverTest(*ctx);
};
